#
# Example calls:
#
# Computing SU(6,2):
#    n:=6; Read("3transp.gap");
# Computing - class in O(5,3):
#    n:=205; sign:=-1; Read("3transp.gap");
# Computing + class in O(5,3):
#    n:=205; sign:=+1; Read("3transp.gap");
# and so on... look at file fischer-groups.gap to find out more about
# other possible values.
#
Read("common.gap");

#
# Load group definition
#
Read("fischer-groups.gap");

# Verify that D is a Fischer space, under the assumption that D is a single conjugacy class
Assert(0, Order(d1)=2);
#Assert(0, ForAll(D, d->Order(d1*d) in [1,2,3]));

# The perp of a set of elements in D is the set of all elements commuting with them.
# We don't need it in this file but elsewhere it is handy.
Perp := function(arg)
	if Size(arg) = 1 and IsList(arg[1]) and not IsMatrix(arg[1]) then
		arg := arg[1];
	fi;
	return Filtered(D, d->ForAll(arg, e->IsOne(Comm(d,e))));
end;;


#
# Compute an affine plane, and the subspace spanned by
# all affine planes.
#
Display("Searching an affine plane");
pi := ConstructAffinePlane(Gperm, d1);
Assert(0, pi = fail or Size(pi) = 9);

Display("Constructing subspace W spanned by plane point sums...");
Wbase := ConstructOrbitSpan(Gperm, pi);;

if pi = fail or (Size(D) = Length(Wbase)) then
	Print(" -> dim L = ",  Size(D) - Length(Wbase), "\n");
else
	Print(" -> dim L <= ",  Size(D) - Length(Wbase), "\n");
fi;

V := GF(2)^Size(D);

# We use that Wbase is in semi-echelon form and read of a "nice"
# basis for a complement U of W in V (here "nice" means: subset of
# the original basis of V). We do this by taking a complement of
# the pivots of vectors in Wbase.
# Then we can use CleanRow to map vectors from V to U = V/W:
Ubase_idx := Difference([1..Size(D)], Pivots(Wbase));
Udim := Size(Ubase_idx);
U := GF(2)^Udim;

#
# Compute the subspace spanned by all sets A_d. The dimension of that equals the
# dimension of the quotient A/V, where V is the ideal of all vanishing sets.
#
Display("Compute the subspace spanned by all sets A_d");

# Find the linear subspace(s) spanned by
# the sets A_d = { e \in D | e and d do not commute }
Ad1 := MovedPoints(d1);;	# points not commuting with d1 are moved by d1
Ad_base := ConstructOrbitSpan(Gperm, Ad1);
Print("dim(A/V) = ", Length(Ad_base), "\n");

Display("Compute the subspace spanned by all vanishing sets");
tmp := NullspaceMat(TransposedMat(Vectors(Ad_base)));
vanishing_base := SemiEchelonBasisMutable(tmp);
Print("dim(V) = ", Length(vanishing_base), "\n");

# Check whether the given set is a vanishing set
IsVanishingSet := function(D, set)
	return CleanRow(vanishing_base, ListToCVec(Size(D), set), false, fail);
	#return ForAll(D, d->IsEvenInt(Size(Intersection(MovedPoints(Image(actHom,d)), X))));
end;;
if pi <> fail then
	Print(" Is the plane pi is a vanishing set? ", IsVanishingSet(D, pi), "\n");
fi;

if pi <> fail and Size(D) <> Length(Wbase) then
	Display("Alternate check for dimension of L...");

	# "Hard" check of the dimension of the subspace W resp. of L
#	I := Ideal(A, [ Sum(Basis(A){pi}) ]);
#	Ir := RightIdeal(A, [ Sum(Basis(A){pi}) ]);
#	Il := LeftIdeal(A, [ Sum(Basis(A){pi}) ]);

	# We want to compute the dimension of an ideal in the "transvection algebra"
	# or "3-transposition algebra" spanned by some vector.
	# To do this efficiently, we use fast linear algebra in several steps.

	# Step 1: Map some gens of A to their adjoint matrices.
	# The trick is to take "enough" of them to generate the whole
	# algebra.

	sc := StabChain(Gperm);
	adjs := [];
	for i in [1..10] do
		j := Random([1..Size(D)]);
		d := d1 ^ (InverseRepresentative(sc, j)^-1);

		# Compute the right adjoint matrix of d
		Print("   Computing adjoint matrix ", i, " for point ", j, "\n");
		m := CVEC_ZeroMat( Size(D), Size(D), 2, 1 );
		for k in [1..Size(D)] do
			#e := GetTransvectionWithNumber(Gperm, d1, k);
			e := d1 ^ (InverseRepresentative(sc, k)^-1);
			if d*e <> e*d then
				l := FindCenterOfTransvection(Gperm, d^e);
				m[k][j] := 1;
				m[k][k] := 1;
				m[k][l] := 1;
			fi;
		od;
		Add(adjs, m);
	od;

	# Step 2: Compute the orbit of the vector pi under the
	# matrices in adjs
	Ibase := ConstructOrbitSpan_for_adj_mats(Gperm, adjs, ListToCVec(Size(D), pi));

	Print(" real dim(W) = ", Length(Ibase), "\n");
	Print(" real dim(L) = ", Length(D) - Length(Ibase), "\n");
fi;
