# We use the cvec package for speed up in the vector processing
if true <> LoadPackage("cvec") then
	Error("Failed to load cvec");
fi;

IsTransvection := function(G, g)
	return 1=RankMat(g-One(G));;
end;;

# Compute all transvection of order 2
Transvections := function(G)
	local g;
	# We could use ConjugacyClasses but that's way overkill for big n.
	# So instead we search a single transvection, "Las Vegas style",
	#  and take its conjugacy class.
	# This of course uses/requires that all transvections are conjugate.
	repeat
		# Find an involution
		repeat g:=PseudoRandom(G); until IsEvenInt(Order(g));
		g:=g^(Order(g)/2);
	until IsTransvection(G, g);
	return Elements(ConjugacyClass(G, g));
end;;

VecToList := function (v)
	return Filtered([1..Size(v)],i->IsOne(v[i]));
end;;

# Turn a list of indices i_n into the vector e_{i_1} + ... + e_{i_n}
ListToVec := function(n, l)
	local i, v;
	v := List([1..n], i->Zero(GF(2)));
	for i in l do
		v[i] := One(GF(2));
	od;
	ConvertToVectorRep(v, 2);
	return v;
end;;

# Turn a list of indices i_n into the algebra element e_{i_1} + ... + e_{i_n}
ListToAlg := function(A, l)
	local B;
	if Length(l) = 0 then
		return Zero(A);
	fi;
	B := BasisVectors(Basis(A));
	return Sum(List(l, i->B[i]));
	#return Sum(B{l});
end;;

# Turn a list of indices i_n into the compressed vector e_{i_1} + ... + e_{i_n}
ListToCVec := function(n, l)
	local i, v;
	v := CVec(CVecClass(2,1,n));
	for i in l do
		v[i] := 1;
	od;
	return v;
end;;

CVecToList := function(v)
#	return Filtered([1..Length(v)], i -> not IsZero(v[i])); 
	return Filtered([PositionNonZero(v)..Length(v)], i -> not IsZero(v[i])); 
#	return Filtered([PositionNonZero(v)..PositionLastNonZero(v)], i -> not IsZero(v[i])); 
end;;


# Determine to which point p1 the transvection d1 corresponds. For this we
# use that the centralizer of d1 equals the point stabilizer of p1.
FindCenterOfTransvection := function(Gperm, d1)
	local tmp, d2, g, n, z;
	# TODO: Slow
	if false then
		# Brute force: Compute the full centralizer
		tmp := Difference(MovedPoints(Gperm), MovedPoints(Centralizer(Gperm,d1)));
	else
		# We avoid computing the full centralizer and instead just try to look at
		# 'sufficiently many' elements of it. We use the 'dihedral group trick' for
		# this: We know that d1 is an involution. Hence for any g in Gperm, d2:=d1^g
		# is also an involution, and so <d1, d2> generate a dihedral group.
		# If n:=Order(d1*d2) is even, then (d1*d2)^(n/2) centralizes d1 and d2.
		# Else, g * (d1*d2)^((n-1)/2) centralizes d1.
		#
		# Actually, since we are in a 3-transposition group, the only
		# possible values for n are 1 (then g is in the centralizer)
		# or 2 and 3. But we keep the method generic anyway :).
		Assert(0, Order(d1)=2);
		tmp := MovedPoints(Gperm);;
		while Length(tmp) > 1 do
			g := PseudoRandom(Gperm);
			d2 := d1^g;
			n := Order(d1*d2);
			if IsEvenInt(n) then
				z := (d1*d2)^(n/2);
			else
				z := g * (d1*d2)^((n-1)/2);
			fi;
			Assert(0, d1^z=d1);
			tmp := Difference(tmp, MovedPoints(z));;
		od;
	fi;
	Assert(0, Size(tmp) = 1);
	return tmp[1];
end;;

# Compute the j-th transvection
GetTransvectionWithNumber := function (Gperm, d1perm, j)
	local sc, p1, g;
	sc := StabChain(Gperm);
	p1 := FindCenterOfTransvection(Gperm, d1perm);
	g := InverseRepresentative(sc, p1) * InverseRepresentative(sc, j)^-1;;
	d1perm := d1perm ^ g;
	Assert(0, j = FindCenterOfTransvection(Gperm, d1perm));
	return d1perm;
end;;

# Find any affine plane, given the set D of transvections, as well actHom,
# the homomorphism from G into the action rep of G on D.
ConstructAffinePlane := function(Gperm, d1)
	local p1, p2, pts1, l1, g, d2, pi, tmp, tmp_pt;

	if NrMovedPoints(Gperm) < 9 then
		return fail;
	fi;

	# Determine to which point p1 the transvection d1 corresponds
	p1 := FindCenterOfTransvection(Gperm, d1);

	# We want to find an affine plane containing the point p1.
	# Find all points collinear to it -- these correspond to transvections
	# *not* commuting with d1, i.e., all elements in D moved by d1.
	pts1 := MovedPoints(d1);;

	# To proceed, we naively would pick an arbitrary line through p1. Then
	# we'd look for a second line through p1 with the property that its points
	# are collinear with the points on the first line.
	# Getting the full affine plane from there then is a simple matter.
	#
	# This approach works if one can iterate over the conjugacy class of d1.
	# But in general, this is a big set and we want to avoid iterating it.
	#
	# So, we work like this: We randomly choose some group element g which 
	# moves p1 to a point p2 different from but collinear to p1. Using g we
	# can now compute all lines through p2 as well as the third point p3 on
	# the line [p1,p2,p3], and all lines through it. Now we just need to find
	# points collinear to all three points, and keep taking linear spans.

	# Find suitable p2 and g such that p2=p1^g
	repeat g := PseudoRandom(Gperm); until p1<>p1^g and p1^g in pts1;
	p2 := p1^g;
	l1 := Set([p1, p2, p2^d1]);
	d2 := d1^g; # This involution fixes the point p2 and interchanges p1 and p2^d1.

	Assert(0, p1^g = FindCenterOfTransvection(Gperm, d2));
	Assert(0, Size(Set([d1,d2,d1^d2])) = 3);
	
	# Find all points collinear to every point on l1. 
	tmp := Intersection(List([d1,d2,d1^d2],MovedPoints));;
	if Size(tmp) = 0 then
		return fail;
	fi;

	# Start building pi:
	pi := Set(l1);

	# Take some point tmp_pt which is collinear to every point on l1.
	# Form all linear spans with it.
	tmp_pt := tmp[1];
	UniteSet(pi, [p1, tmp_pt, tmp_pt^d1]);	# d1 is the involution with center p1
	UniteSet(pi, [p2, tmp_pt, tmp_pt^d2]);	# d2 is the involution with center p2
	UniteSet(pi, [p2^d1, tmp_pt, ((tmp_pt^d2)^d1)^d2]);	# d1^d2=d2^d1 is the involution with center p1^d2=p2^d1
	
	# Now add all spans with "new" points in pi
	tmp := Difference(pi, l1);
	for tmp_pt in tmp do
		UniteSet(pi, [p1, tmp_pt, tmp_pt^d1]);
		UniteSet(pi, [p2, tmp_pt, tmp_pt^d2]);
		UniteSet(pi, [p2^d1, tmp_pt, ((tmp_pt^d2)^d1)^d2]);
	od;
	Assert(0, Size(pi) = 9);
	return pi;
end;;

GFactors := function (y)
	return List(MTX.CollectedFactors(y), f->[f[1].field^f[1].dimension, f[2]]);
end;;

# Given a set D of transvection, compute an algebra on 2^D with
# the following multiplication:
# We have x*y = x+y+z = x+y+yxy = x+y+xyx if x,y do not commute,
#  else x*y=0  (i.e. if x,y commute)
TransvectionAlgebra := function (D)
	local T, i, j, d, e, k;
	T := EmptySCTable(Size(D), Zero(GF(2)), "symmetric");;
	for i in [1..Size(D)] do for j in [1..Size(D)] do
		d := D[i]; e := D[j];
		if d*e <> e*d then
			k := Position(D, e*d*e);
			SetEntrySCTable(T, i, j, [1, i, 1, j, 1, k]);
		fi;
	od; od;
	return AlgebraByStructureConstants(GF(2), T);
end;;

# Symmetric difference of two sets
SymDiff := function(A,B)
	return Difference(Union(A,B), Intersection(A,B));
end;;

# Compute whether a given 'space' (= subset of D) is closed
# under linear spans.
IsClosedSpace := function(spc)
	return ForAll(spc, x->ForAll(spc, y->x^y in spc));;
end;;

# Inefficient way to compute the closure of a 'space' (= subset of D)
ClosureSpace := function(spc)
	local tmp, x, y;
	spc := Set(spc);
	repeat
		tmp := ShallowCopy(spc);
		for x in tmp do
			for y in tmp do
				AddSet(spc, x^y);
			od;
		od;
	until spc = tmp;
	return spc;
end;;

# Given a set pi of indices into D, compute the subspace of V=2^D
# spanned by the pointsum of all elements in the G-orbit of pi.
#
# We use this to compute the ideal spanned by all affine plane pointsum.
# But it can also be used for any other sets, e.g. octas, singular lines, ...
GrowOrbitSpan_FAST_HI_MEM := function(Gperm, pi, Wbase)
	local n, sp, top, debug_count, g, cands, vPi, v;
	n := NrMovedPoints(Gperm);
	if pi <> fail then
		debug_count := 0;
		CleanRow(Wbase, ListToCVec(n, pi), true, fail);
		cands := [ pi ];
		sp := 1;
		while sp > 0 do
			top := cands[sp];
			sp := sp - 1;	# pop front element from stack
			for g in GeneratorsOfGroup(Gperm) do
				vPi := OnSets(top, g);
				v := ListToCVec(n, vPi);
				if not CleanRow(Wbase, v, true, fail) then
					sp := sp + 1;
					if sp > Size(cands) then
						cands[2*Size(cands)] := 0;
					fi;
					cands[sp] := vPi;
				fi;

				debug_count := debug_count + 1;
				if 0 = debug_count mod 100 then
					Print(debug_count, " (queue: ", sp, ")\n");
				fi;
				if 0 = debug_count mod 1000 then
					Print("  dim(Wbase) = ", Length(Wbase), "\n");
				fi;
			od;
		od;
		# TODO: Turn these Prints info Infos
		Print("  dim(W) = ", Length(Wbase), "\n");
	fi;
	
	return Wbase;
end;;

# InstallMethod( Length, "for a semi echelonised basis", [ SEBType ],
#   function(b)
#     return Length(b!.vectors);
#   end );


GrowOrbitSpan_SLOW_LO_MEM := function(Gperm, pi, Wbase)
	local n, i, top, g, vPi, v;
	n := NrMovedPoints(Gperm);
	if pi <> fail then
		CleanRow(Wbase, ListToCVec(n, pi), true, fail);
		i := 1;
		while i <= Length(Wbase!.pivots) do
#			v := Wbase[i];
#			top := Filtered([PositionNonZero(v)..Length(v)], i -> not IsZero(v[i])); 
			top := CVecToList(Wbase!.vectors[i]);
			for g in GeneratorsOfGroup(Gperm) do
				vPi := OnSets(top, g);
				v := ListToCVec(n, vPi);
				CleanRow(Wbase, v, true, fail);

			od;
			if 0 = i mod 100 then
				Print(" pos: ", i, " / ", Length(Wbase!.pivots), "\n");
			fi;
			i := i + 1;
		od;
		# TODO: Turn these Prints info Infos
		Print(" End result: dim(W) = ", Length(Wbase!.pivots), "\n");
	fi;
	
	return Wbase;
end;;

#GrowOrbitSpan := GrowOrbitSpan_FAST_HI_MEM;
GrowOrbitSpan := GrowOrbitSpan_SLOW_LO_MEM;

ConstructOrbitSpan := function(Gperm, pi)
	return GrowOrbitSpan(Gperm, pi, EmptySemiEchelonBasis(CMat([], CVecClass(2,1,NrMovedPoints(Gperm)))));
end;;




GrowOrbitSpan_SLOW_LO_MEM_for_adj_mats := function(adjs, pi_vec, Wbase)
	local i, top, g, v;
#	if pi <> fail then
		CleanRow(Wbase, pi_vec, true, fail);
		i := 1;
		while i <= Length(Wbase!.pivots) do
#			v := Wbase[i];
#			top := Filtered([PositionNonZero(v)..Length(v)], i -> not IsZero(v[i])); 
			top := Wbase!.vectors[i];
			for g in adjs do
				v := top * g;
				CleanRow(Wbase, v, true, fail);

			od;
			if 0 = i mod 100 then
				Print(" pos: ", i, " / ", Length(Wbase!.pivots), "\n");
			fi;
			i := i + 1;
		od;
		# TODO: Turn these Prints info Infos
		Print(" End result: dim(W) = ", Length(Wbase!.pivots), "\n");
#	fi;
	
	return Wbase;
end;;

GrowOrbitSpan_for_adj_mats := GrowOrbitSpan_SLOW_LO_MEM_for_adj_mats;

ConstructOrbitSpan_for_adj_mats := function(Gperm, adjs, pi_vec)
	local tmp;
	tmp := EmptySemiEchelonBasis(CMat([], CVecClass(2,1,NrMovedPoints(Gperm))));
	return GrowOrbitSpan_for_adj_mats(adjs, pi_vec, tmp);
end;;
